/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia
  
  The DXC Framework is free software: you can redistribute it and/or modify it 
  under the terms of the GNU Lesser General Public License as published by the 
  Free Software Foundation, either version 3 of the License, or (at your 
  option) any later version. 

  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
  details. 

  You should have received a copy of the GNU Lesser General Public License 
  along with this program. If not, see <http://www.gnu.org/licenses/> 
 */
package org.dxc.api.connection;

import org.dxc.api.DxcCallback;
import org.dxc.api.datatypes.DxcData;
import org.dxc.api.connection.ConnectionCallback;

import java.io.IOException;
import java.net.Socket;
import java.net.ServerSocket;
import java.util.LinkedList;

/**
 * Interface for sending and receiving DxcData.
 * 
 * A DxcConnector is created via the appropriate call to a ConnectorFactory.
 */
public abstract class DxcConnector {

    protected DxcCallback callback;
    private String source = null;
    private LinkedList<Thread> threads;
    private LinkedList<Socket> sockets;
    private LinkedList<ServerSocket> ssockets;

    public DxcConnector(DxcCallback c) {
        this.callback = c;
        this.threads = new LinkedList<Thread>();
        this.sockets = new LinkedList<Socket>();
        this.ssockets = new LinkedList<ServerSocket>();
    }

    public DxcConnector(DxcCallback c, String source) {
        this.callback = c;
        this.source = source;
        this.threads = new LinkedList<Thread>();
        this.sockets = new LinkedList<Socket>();
        this.ssockets = new LinkedList<ServerSocket>();
    }

    public void shutdown() {
        for (ServerSocket s : ssockets) {
            try {
                s.close();
            } catch (Exception ignore) {
            }
        }

        for (Socket s : sockets) {
            try {
                s.close();
            } catch (Exception ignore) {
            }
        }

        for (Thread t : threads) {
            try {
                t.join();
            } catch (InterruptedException ignore) {
            }
        }
    }

    /**
     * Sends the DxcData message to the appropriate recipient.
     */
    public abstract void sendMessage(DxcData d);

    protected Socket connect(String host, String port) throws IOException {
        return connect(host, Integer.valueOf(port));
    }

    protected Socket connect(String host, int port) throws IOException {
        Socket s = new Socket(host, port);
        sockets.add(s);
        Thread t;
        if (source == null)
            t = new Thread(new StreamListener(callback, s.getInputStream()));
        else
            t = new Thread(new StreamListener(callback, s.getInputStream(),
                    source));

        threads.add(t);
        t.start();
        return s;
    }

    protected void listen(String port, ConnectionCallback c) throws IOException {
        listen(Integer.valueOf(port), c);
    }

    protected void listen(int port, ConnectionCallback c) throws IOException {
        ServerSocket s = new ServerSocket(port);
        ssockets.add(s);

        new Thread(new ServerSocketMonitor(this.callback, s, c)).start();
    }
}