/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia
  
  The DXC Framework is free software: you can redistribute it and/or modify it 
  under the terms of the GNU Lesser General Public License as published by the 
  Free Software Foundation, either version 3 of the License, or (at your 
  option) any later version. 

  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
  details. 

  You should have received a copy of the GNU Lesser General Public License 
  along with this program. If not, see <http://www.gnu.org/licenses/> 
 */
package org.dxc.api.datatypes;

import java.io.OutputStream;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.List;

/**
 * Defines the data structure for a test article's Diagnosis.
 * 
 * A diagnosis is a set of candidates, weighed by their relative probability.
 * The weights of these candidates are normalized upon construction of the
 * datatype, resulting in weights that yield 1 when added together. An
 * additional detection signal and isolation signal indicate whether the
 * reasoner has detected something wrong with the system (detectionSignal =
 * true) and wether it has isolated the fault to specific component failure
 * modes (isolationSignal = true). If the isolationSignal is false, the
 * Candidate set, regardless of its contents, may be effectively treated as
 * empty.
 */
public class DiagnosisData extends DxcData {
    private CandidateSet candidateSet;
    private boolean detectionSignal;
    private boolean isolationSignal;
    private String notes;

    /**
     * Constructor to initialize the DiagnosisData.
     * 
     * @param detectionSignal
     *            Set to true if the system is believed to be in a faulty state,
     *            even if there is not yet an isolation.
     * @param isolationSignal
     *            Set to true if faults have been isolated, ie. candidates
     *            exist.
     * @param candidateSet
     *            A set of Candidates (hypotheses). Their weights will be
     *            normalized, so that they all add up to one.
     * @param notes
     *            A string for diagnosis algorithms to supply additional
     *            information for their own purposes.
     */
    public DiagnosisData(boolean detectionSignal, boolean isolationSignal,
            CandidateSet candidateSet, String notes) {
        super();
        this.detectionSignal = detectionSignal;
        this.isolationSignal = isolationSignal;
        this.candidateSet = candidateSet;
        this.notes = (notes == null) ? "" : notes;

        if (this.candidateSet == null) {
            this.candidateSet = new CandidateSet();
        }

        normalizeWeights();
    }

    /**
     * Constructor to initialize the DiagnosisData.
     * 
     * @param detectionSignal
     *            Set to true if the system is believed to be in a faulty state,
     *            even if there is not yet an isolation.
     * @param isolationSignal
     *            Set to true if faults have been isolated, ie. candidates
     *            exist.
     * @param candidateSet
     *            A set of Candidates (hypotheses). Their weights will be
     *            normalized, so that they all add up to one.
     * @param notes
     *            A string for diagnosis algorithms to supply additional
     *            information for their own purposes.
     */
    public DiagnosisData(long timestamp, boolean detectionSignal,
            boolean isolationSignal, CandidateSet candidateSet, String notes) {
        super(timestamp);
        this.detectionSignal = detectionSignal;
        this.isolationSignal = isolationSignal;
        this.candidateSet = candidateSet;
        this.notes = (notes == null) ? "" : notes;

        if (this.candidateSet == null) {
            this.candidateSet = new CandidateSet();
        }

        normalizeWeights();
    }

    // this constructor is meant to be called from the parser
    public DiagnosisData(long timestamp, boolean detectionSignal,
            boolean isolationSignal, List<FaultSet> candidates,
            List<Double> weights, String notes) {
        super(timestamp);
        this.detectionSignal = detectionSignal;
        this.isolationSignal = isolationSignal;
        this.candidateSet = new CandidateSet();
        this.notes = (notes == null) ? "" : notes;

        if (candidates != null) {
            Iterator<FaultSet> ci = candidates.iterator();

            while (ci.hasNext()) {
                Candidate c = new Candidate(ci.next());

                if (!weights.isEmpty()) {
                    c.setWeight(weights.remove(0));
                }

                this.candidateSet.add(c);
            }

            normalizeWeights();
        }
    }

    /**
     * True if, according to the diagnosis, the system is believed to be in a
     * faulty state.
     */
    public boolean getDetectionSignal() {
        return detectionSignal;
    }

    /**
     * True if faults have been isolated, ie. candidates exist.
     */
    public boolean getIsolationSignal() {
        return isolationSignal;
    }

    /**
     * Returns test-article-specific notes.
     */
    public String getNotes() {
        return notes;
    }

    /**
     * Returns the set of candidates.
     */
    public java.util.Set<Candidate> getCandidateSet() {
        return candidateSet;
    }

    public boolean equals(Object o) {
        if (o == this)
            return true;
        if ((o == null) || (o.getClass() != this.getClass()))
            return false;

        DiagnosisData d = (DiagnosisData) o;
        if ((d.getTimeStamp() != getTimeStamp())
                || (!d.getSource().equals(getSource()))
                || (d.detectionSignal != detectionSignal)
                || (d.isolationSignal != isolationSignal)
                || (!d.candidateSet.equals(candidateSet))
                || (!d.notes.equals(notes)))
            return false;

        return true;
    }

    public int hashCode() {
        int hash = (int) getTimeStamp() + (detectionSignal ? 103 : 53);
        hash = 31 * hash + getSource().hashCode();
        hash = 31 * hash + (isolationSignal ? 1001 : 503);
        hash = 31 * hash + candidateSet.hashCode();
        hash = 31 * hash + notes.hashCode();
        return hash;
    }

    public OutputStream put(OutputStream os) {
        PrintStream out = new PrintStream(os);
        out.append("diagnosis @");
        out.append(Long.toString(this.getTimeStamp()));
        if (this.getSource() != null)
            out.append(" <" + this.getSource() + ">");
        out.append(" detection = " + (this.detectionSignal ? "true" : "false")
                + ", ");
        out.append("isolation = " + (this.isolationSignal ? "true" : "false")
                + ", ");

        out.append("candidates = ");

        out.append(candidateSet.toString());

        out.append(", notes = '");
        out.append(notes + "';\n");

        return out;
    }

    private void normalizeWeights() {
        if (candidateSet.isEmpty())
            return;

        double total = 0;
        Iterator<Candidate> i = this.candidateSet.iterator();

        while (i.hasNext()) {
            Candidate c = i.next();
            if (c.getWeight() < 0)
                c.setWeight(-c.getWeight());
            total += c.getWeight();
        }

        i = this.candidateSet.iterator();
        while (i.hasNext()) {
            Candidate c = i.next();
            c.setWeight(c.getWeight() / total);
        }
    }
}
