package org.dxc.api.datatypes;

import java.io.File;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;

import org.dxc.api.DxcConfig;
import org.w3c.dom.*;

/**
 * System catalog parser.
 * 
 * Provides a data structure for accessing information contained in a system's
 * XML catalog.
 */
public class SystemCatalogXML implements SystemCatalog {
    private static Map<String, SystemCatalog> catMap;
    private Map<String, String> compIdToType;
    private Map<String, Set<String>> typeToModes;
    private Map<String, Set<String>> typeToFaultModes;
    private Map<String, Double> sensorCosts;
    private String name;
    private int numModes;
    private int numFaultModes;
    private int nomComponents;

    private SystemCatalogXML(String system) throws Exception {
        name = system;
        compIdToType = new HashMap<String, String>();
        typeToModes = new HashMap<String, Set<String>>();
        typeToFaultModes = new HashMap<String, Set<String>>();
        sensorCosts = new HashMap<String, Double>();
        numModes = 0;
        numFaultModes = 0;
        nomComponents = 0;
        String sysDir = DxcConfig.getInstance().get("SYSTEMCAT_DIR");
        File f = new File(sysDir + "/" + system + ".xml");

        DocumentBuilder db = DocumentBuilderFactory.newInstance()
                .newDocumentBuilder();
        Document doc = db.parse(f);
        doc.getDocumentElement().normalize();

        // get component type -> modes map

        Map<String, Set<String>> typeMdRefMap = new HashMap<String, Set<String>>();
        Map<String, Double> typeProbeCostMap = new HashMap<String, Double>();
        Map<String, Set<String>> mdRefModesMap = new HashMap<String, Set<String>>();
        Map<String, Set<String>> mdRefFaultModesMap = new HashMap<String, Set<String>>();

        NodeList nl = doc.getElementsByTagName("componentTypeCatalog").item(0)
                .getChildNodes();
        Set<String> sensorTypes = new HashSet<String>();

        for (int i = 0; i < nl.getLength(); i++) {
            Node nm = (Node) nl.item(i);
            if (nm.getNodeType() == Node.ELEMENT_NODE) {
                String componentType = getChildByName(nm, "name")
                        .getFirstChild().getNodeValue();

                NamedNodeMap att = nm.getAttributes();
                Node typeN = att.getNamedItem("xsi:type");
                if (typeN != null) {
                    String type = typeN.getNodeValue();

                    if (type.equals("sensor")) {
                        sensorTypes.add(componentType);
                        Node probeCost = getChildByName(nm, "probeCost");
                        if (probeCost != null) {
                            Double d = Double.parseDouble(probeCost
                                    .getFirstChild().getNodeValue());
                            typeProbeCostMap.put(componentType, d);
                        }
                    }
                }

                Node mr = getChildByName(nm, "modesRef");

                while (mr != null) {
                    String modesRef = mr.getFirstChild().getNodeValue();

                    Set<String> mdRefs = typeMdRefMap.get(componentType);
                    if (mdRefs == null) {
                        mdRefs = new HashSet<String>();
                    }
                    mdRefs.add(modesRef);
                    typeMdRefMap.put(componentType, mdRefs);

                    mr = getNextSiblingByName(mr, "modesRef");
                }
            }
        }

        nl = doc.getElementsByTagName("modeGroup");

        for (int i = 0; i < nl.getLength(); i++) {
            Node nm = (Node) nl.item(i);
            if (nm.getNodeType() == Node.ELEMENT_NODE) {
                Node mgNode = getChildByName(nm, "name");
                String modeGroup = mgNode.getFirstChild().getNodeValue();
                Set<String> modes = new HashSet<String>();
                Set<String> faultModes = new HashSet<String>();

                Node md = mgNode.getNextSibling();

                while (md != null) {
                    if (md.getNodeName().equals("mode")) {
                        NamedNodeMap att = md.getAttributes();
                        Node typeN = att.getNamedItem("xsi:type");
                        String type = typeN.getNodeValue();
                        String s = getChildByName(md, "name").getFirstChild()
                                .getNodeValue();

                        modes.add(s);

                        if (type.equals("faultMode")) {
                            faultModes.add(s);
                        }
                    }
                    md = md.getNextSibling();
                }

                mdRefModesMap.put(modeGroup, modes);
                mdRefFaultModesMap.put(modeGroup, faultModes);
            }
        }

        Iterator<String> j = typeMdRefMap.keySet().iterator();
        while (j.hasNext()) {
            String type = j.next();
            Set<String> modeRefs = typeMdRefMap.get(type);

            Set<String> modes = new HashSet<String>();
            Set<String> faultModes = new HashSet<String>();
            for (String s : modeRefs) {
                if (mdRefModesMap.containsKey(s))
                    modes.addAll(mdRefModesMap.get(s));
                if (mdRefFaultModesMap.containsKey(s))
                    faultModes.addAll(mdRefFaultModesMap.get(s));
            }

            typeToModes.put(type, modes);
            typeToFaultModes.put(type, faultModes);

        }

        nl = doc.getElementsByTagName("component");

        // get component ids

        for (int i = 0; i < nl.getLength(); i++) {
            Node nm = (Node) nl.item(i);
            String componentID = getChildByName(nm, "name").getFirstChild()
                    .getNodeValue();
            String componentType = getChildByName(nm, "componentType")
                    .getFirstChild().getNodeValue();

            if (sensorTypes.contains(componentType)) {
                try {
                    Node pc = getChildByName(nm, "probeCost");
                    if (pc != null) {
                        String pcost = pc.getFirstChild().getNodeValue();
                        sensorCosts.put(componentID, Double.valueOf(pcost));
                    } else if (typeProbeCostMap.containsKey(componentType)) {
                        sensorCosts.put(componentID, typeProbeCostMap
                                .get(componentType));
                    }
                } catch (Exception ex) {
                    sensorCosts.put(componentID, 0.0);
                }
            }

            compIdToType.put(componentID, componentType);
            Set<String> modes = typeToModes.get(componentType);
            Set<String> faultModes = typeToFaultModes.get(componentType);
            numModes += modes.size();
            if (faultModes.size() == 0)
                nomComponents++;
            else
                numFaultModes += faultModes.size();
        }

        db.reset();

    }

    /**
     * Get a system's catalog.
     * 
     * @param system
     *            the system's name
     * 
     *            System catalogs are searched for in $DXC_HOME/[SYSTEMCAT_DIR]
     *            (see DXC.cfg for SYSTEMCAT_DIR).
     */
    public static SystemCatalog getInstance(String system) throws Exception {
        if (catMap == null) {
            catMap = new HashMap<String, SystemCatalog>();
        }

        if (catMap.containsKey(system)) {
            return catMap.get(system);
        } else {
            SystemCatalog s = new SystemCatalogXML(system);
            catMap.put(system, s);
            return s;
        }
    }

    /**
     * The name of the system.
     */
    public String getSystemName() {
        return name;
    }

    /**
     * The set of component IDs the system contains.
     */
    public Set<String> getComponentIDs() {
        return compIdToType.keySet();
    }

    public Set<String> getModes(String componentID) {
        String type = compIdToType.get(componentID);

        if (type == null)
            return null;

        return typeToModes.get(type);
    }

    /**
     * The set of mode IDs of a given component
     * 
     * @param componentID
     *            component's ID
     */
    public Set<String> getFaultModes(String componentID) {
        String type = compIdToType.get(componentID);

        if (type == null)
            return null;

        return typeToFaultModes.get(type);
    }

    /**
     * Test the validity of a componentID.
     * 
     * @param componentID
     *            component's ID
     */
    public boolean isComponentValid(String componentID) {
        if (compIdToType.keySet().contains(componentID))
            return true;

        return false;
    }

    public boolean isModeValid(String componentID, String modeID) {
        if (compIdToType.keySet().contains(componentID)) {
            String type = compIdToType.get(componentID);
            if (this.typeToFaultModes.get(type).contains(modeID))
                return true;
        }

        return false;
    }

    /**
     * Given a valid component, test the validity of a mode ID.
     * 
     * @param componentID
     *            component ID
     * @param modeID
     *            mode ID
     */
    public boolean isFaultModeValid(String componentID, String modeID) {
        if (compIdToType.keySet().contains(componentID)) {
            String type = compIdToType.get(componentID);
            if (this.typeToFaultModes.get(type).contains(modeID))
                return true;
        }

        return false;
    }

    /**
     * Test the validity of a Fault.
     */
    public boolean isFaultModeValid(Fault f) {
        if (compIdToType.keySet().contains(f.getComponentID())) {
            String type = compIdToType.get(f.getComponentID());
            if (this.typeToFaultModes.get(type).contains(f.getModeID()))
                return true;
        }

        return false;
    }

    public int getNumModes() {
        return numModes;
    }

    public int getNumFaultModes() {
        return numFaultModes;
    }

    public int getNumNoFaultComponents() {
        return nomComponents;
    }

    private Node getChildByName(Node n, String name) {
        NodeList nl = n.getChildNodes();
        for (int i = 0; i < nl.getLength(); i++) {
            Node ret = nl.item(i);
            if (ret.getNodeName().equals(name))
                return ret;
        }
        return null;
    }

    private Node getNextSiblingByName(Node n, String name) {
        Node s = n.getNextSibling();
        while (s != null) {
            if (s.getNodeName().equals(name))
                return s;
            s = s.getNextSibling();
        }
        return null;
    }

    public Map<String, Double> getSensorsAndCosts() {
        return sensorCosts;
    }
}
