/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
/**
 * ExampleDA demonstrates how a diagnosis algorithm should
 * send and receive scenario data.
 */

#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include <iostream>

#include <DxcCallback.hpp>
#include <DxcDataTypes.hpp>
#include <Connection/ConnectorFactory.hpp>

using namespace Dxc;
using namespace std;

static DxcConnector *connector;
static bool run = true;

class ExampleDA
{
public:
    void sendDiagnosis();
};

void ExampleDA::sendDiagnosis()
{
    // In this example we send a diagnosis for every 10th received SensorData.
    static int i = -1;
    i++;
    if (i % 10 != 0) return;

    CandidateSet exampleCandidateSet;
    Candidate c1, c2, c3;

    /* Here we populate candidates with made-up data.
     * In reality component IDs and mode IDs must match
     * XML system descriptions.
     */
    c1.getFaults()->insert(Fault("c", "stuckOpen"));
    c1.getFaults()->insert(Fault("d", "stuckClosed"));
    c1.setWeight(0.2);

    c2.getFaults()->insert(Fault("a", "bias"));
    c2.getFaults()->insert(Fault("b", "switchedOff"));
    c2.setWeight(0.3);

    c3.getFaults()->insert(Fault("c", "stuckClosed"));
    c3.getFaults()->insert(Fault("d", "stuckOpen"));
    c3.setWeight(0.5);

    exampleCandidateSet.insert(c1);
    exampleCandidateSet.insert(c2);
    exampleCandidateSet.insert(c3);

    DiagnosisData* exampleDiagnosis = new DiagnosisData(true, true,
            exampleCandidateSet, "example diagnosis");

    connector->sendMessage(exampleDiagnosis);
}

ExampleDA *exampleDA = NULL;

/**
 * This is the DA's implementation of the DxcCallback interface.
 * After an instance of this class is registered with a DxcConnector,
 * the connector will execute its processData() method every time
 * data is received.
 */
class ExampleCallback : public DxcCallback
{
private:
    virtual void processData(const DxcData * msg)
    {
        /* msg is incoming data. Find out what type it is,
         * downcast it to that DxcData type, and process accordingly.
         */
        if (string(typeid(*msg).name()) == string(typeid(SensorData).name()))
        {
            // update observations
            const SensorData* sd = dynamic_cast<const SensorData*>(msg);
            SensorData::SensorValueMap::const_iterator i =
                sd->getSensors()->begin();

            for (; i != sd->getSensors()->end(); i++)
            {
                string sensorID = i->first;
                const Value* v = i->second;

                /* Here we don't know what datatype the value is, but this
                 * information can be looked up by sensor ID in the system's
                 * XML catalog (TinyXML, a lightweight parser, is included
                 * in <Parser/tinyxml.hpp>)
                 */
                if (string(typeid(*v).name()) ==
                        string(typeid(RealValue).name()))
                {
                    // cast to RealValue, retreive value
                    double dv = dynamic_cast<const RealValue*>(v)->get();
                    // then, write (sensorID, sv) to DA model/data
                    // structure for diagnosis
                }
                else if (string(typeid(*v).name()) ==
                         string(typeid(BoolValue).name()))
                {
                    // cast to boolean, et cetera
                }
            }

            exampleDA->sendDiagnosis();
        }
        else if (string(typeid(*msg).name()) ==
                 string(typeid(CommandData).name()))
        {
            const CommandData* cd = dynamic_cast<const CommandData*>(msg);
            CommandSet::const_iterator i = cd->getCommands()->begin();

            for (; i != cd->getCommands()->end(); i++)
            {
                string commandID = i->getCommandID();
                const Value* val = i->getValue();
                // update expected system state
            }
        }
        else if (string(typeid(*msg).name()) ==
                 string(typeid(ScenarioStatusData).name()))
        {
            // check for scenario end, exit if ended
            const ScenarioStatusData* stat =
                dynamic_cast<const ScenarioStatusData *>(msg);

            if (stat->getStatus() == ScenarioStatusData::SDS_ENDED)
            {
                run = false;
            }
        }
        else if (typeid(msg).name() == typeid(ErrorData).name())
        {
            cout << "DA received error: " << msg << endl;
        }
    };
};

int main(int argc, char **argv)
{
    /* Create our implementation of DxcCallback and pass it to a new
     * DxcConnector. A scenario data source (SDS) must be running
     * in order to connect successfully.
     */
    ExampleCallback callback;

    try
    {
        connector = ConnectorFactory::getDAConnector(&callback);
    }
    catch (...)
    {
        cerr << "DA unable to connect. Make sure ScenarioLoader " <<
             "(or StandaloneSDS) is running." << endl;
        return 1;
    }

    exampleDA = new ExampleDA;

    // signal to the system that we're prepared to receive data
    ScenarioStatusData ssd(ScenarioStatusData::DA_READY);
    connector->sendMessage(&ssd);

    while (run)
    {
        msleep(100);
    }

    delete exampleDA;

    // clean up after yourself!
    delete connector;

    return 0;
}
