/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#include <stdlib.h>
#endif

#include <iostream>

#include <DxcDataTypes.hpp>
#include <DataTypes/SystemCatalog.hpp>
#include <Connection/ConnectorFactory.hpp>

#include "Parser/MessageDriver.hpp"

#ifdef _WIN32
#include <Pthreads-w32/pthread.h>
#define pthreadsInit pthread_win32_process_attach_np
#define pthreadsStop pthread_win32_process_detach_np
#define timeval Dxc::timeval
#else
#include <pthread.h>
#include <sys/time.h>
#define pthreadsInit void
#define pthreadsStop void
#endif
#include <time.h>
#include <errno.h>
#include <sys/timeb.h>

using namespace Dxc;
using namespace std;

static DxcConnector *connector;
static set<string>* freeSensors = NULL;
static const SensorData* lastSensors = NULL;
static pthread_mutex_t lastSensorsMutex;
static pthread_mutex_t connectorMutex;
static pthread_mutex_t daMutex;
static pthread_cond_t daCond;
static bool abbrev = false;
static ostream* output = NULL;
static const SystemCatalog* sd = NULL;

static unsigned int transFailedMessages = 0;

class SDSCallback : public DxcCallback
{
private:
    virtual void processData(const DxcData *msg)
    {
        if (string(typeid(*msg).name()) == string(typeid(ScenarioStatusData).name()))
        {
            const ScenarioStatusData* stat = dynamic_cast<const ScenarioStatusData *>(msg);

            if (stat->getStatus() == ScenarioStatusData::SR_READY)
            {
                ScenarioStatusData ssd(ScenarioStatusData::SDS_AWAITING_DA);
                pthread_mutex_lock(&connectorMutex);
                connector->sendMessage(&ssd);
                pthread_mutex_unlock(&connectorMutex);
            }
            else if (stat->getStatus() == ScenarioStatusData::DA_READY)
            {
                ScenarioStatusData ssd(ScenarioStatusData::STARTED);
                pthread_mutex_lock(&connectorMutex);
                connector->sendMessage(&ssd);
                pthread_mutex_unlock(&connectorMutex);

                pthread_mutex_lock(&daMutex);
                pthread_cond_broadcast(&daCond);
                pthread_mutex_unlock(&daMutex);
            }
            else if (stat->getStatus() == ScenarioStatusData::DA_ENDED)
            {
                pthread_mutex_lock(&daMutex);
                pthread_cond_broadcast(&daCond);
                pthread_mutex_unlock(&daMutex);
            }
        }
        else if (string(typeid(*msg).name()) == string(typeid(CommandData).name()))
        {
            const CommandData* cmd = dynamic_cast<const CommandData*>(msg);
            CommandSet::const_iterator i = cmd->getCommands()->begin();

            SensorData::SensorValueMap* nsvm = NULL;
            pthread_mutex_lock(&lastSensorsMutex);
            for (; i != cmd->getCommands()->end(); i++)
            {
                if (i->getCommandID() == "probe")
                {
                    if (nsvm == NULL)
                        nsvm = new SensorData::SensorValueMap();

                    const StringValue* sv(
                        dynamic_cast<const StringValue*>(i->getValue()));

                    if (sv == NULL)
                        return;

                    if (lastSensors != NULL)
                    {
                        const SensorData::SensorValueMap* svm =
                            lastSensors->getSensors();

                        SensorData::SensorValueMap::const_iterator j;
                        j = svm->find(sv->get());
                        if (j != svm->end())
                        {
                            nsvm->insert(make_pair(j->first, j->second));
                        }
                    }
                }
            }
            pthread_mutex_unlock(&lastSensorsMutex);

            if (nsvm != NULL)
            {
                SensorData s(DxcData::currentTime(), *nsvm);
                pthread_mutex_lock(&connectorMutex);
                connector->sendMessage(&s);
                pthread_mutex_unlock(&connectorMutex);
                delete nsvm;
            }
        }
        else if (string(typeid(*msg).name()) == string(typeid(ErrorData).name()))
        {
            pthread_mutex_lock(&connectorMutex);
            if (connector)
            {
                connector->sendMessage(msg);
                exit(0);
            }
            pthread_mutex_unlock(&connectorMutex);
        }
        else if (string(typeid(*msg).name()) == string(typeid(RecoveryData).name()))
        {
            pthread_mutex_lock(&connectorMutex);
            if (connector)
            {
                connector->sendMessage(msg);
            }
            pthread_mutex_unlock(&connectorMutex);
        }

        if (output)
        {
            DxcData *tmp = msg->clone();
            tmp->setTimeStamp(DxcData::currentTime());
            *output << tmp;
        }
    };
};

void validateData(const DxcData* msg)
{
    if (!sd)
    {
        return;
    }

    if ((string)typeid(*msg).name() == (string)typeid(SensorData).name())
    {
        const SensorData* s = dynamic_cast<const SensorData*>(msg);
        const SensorData::SensorValueMap *svm = s->getSensors();;
        SensorData::SensorValueMap::const_iterator i;
        i = svm->begin();

        for (; i != svm->end(); i++)
        {
            if (!sd->isComponentValid(i->first))
            {
                ostringstream oss;
                oss << "SDS: Error: Invalid component id `" << i->first
                    << "' at timestamp " << s->getTimeStamp();
                throw DxcException(oss.str());
            }
        }
    }
    if ((string)typeid(*msg).name() == (string)typeid(FaultInjectData).name())
    {
        const FaultInjectData* fid = dynamic_cast<const FaultInjectData*>(msg);
        Fault f = fid->getFault();

        if (!sd->isComponentValid(f.getComponentID()))
        {
            ostringstream oss;
            oss << "SDS: ERROR: Invalid component id `" << f.getComponentID()
                << "' in faultInjection at timestamp " << fid->getTimeStamp();
            throw DxcException(oss.str());
        }

        if (!sd->isFaultModeValid(f.getComponentID(), f.getModeID()))
        {
            ostringstream oss;
            oss << "SDS: ERROR: Invalid mode id `" << f.getModeID()
                << "' in faultInjection at timestamp " << fid->getTimeStamp();
            throw DxcException(oss.str());
        }
    }
};

DxcData* cloneMessage(DxcData* msg)
{
    if (freeSensors && (string)typeid(*msg).name() == (string)typeid(SensorData).name())
    {
        SensorData* smsg = dynamic_cast<SensorData*>(msg);
        pthread_mutex_lock(&lastSensorsMutex);
        lastSensors = smsg;
        pthread_mutex_unlock(&lastSensorsMutex);
        const SensorData::SensorValueMap *svm = smsg->getSensors();
        SensorData::SensorValueMap newsvm;
        SensorData::SensorValueMap::const_iterator i = svm->begin();
        while (i != svm->end())
        {
            set<string>::iterator j = freeSensors->find(i->first);
            if (j != freeSensors->end())
            {
                newsvm[i->first] = i->second->clone();
            }
            i++;
        }

        return new SensorData(smsg->getTimeStamp(), newsvm);
    }
    else if (string(typeid(*msg).name()) == string(typeid(ScenarioData).name()))
    {
        ScenarioData* s = dynamic_cast<ScenarioData*>(msg->clone());
        string notes = s->getNotes();

        if (sd)
        {
            if (sd->getSystemName() != s->getSystemName())
            {
                ostringstream oss;
                oss << "SDS: ERROR: System name specified via command line " <<
                    "doesn't match ScenarioData system name.";
                throw DxcException(oss.str());
            }
        }
        else
        {
            notes += "SDS: running without validation. ";
        }

        if (abbrev)
            notes += "Running in abbreviated mode.";

        s->setNotes(notes);
        return s;
    }

    return msg->clone();
}

void sendMessage(DxcData *msg)
{
    static Timestamp startTime = -1;
    static Timestamp msgTime0 = msg->getTimeStamp();
    static bool once = true;

    if (once)
    {
        startTime = DxcData::currentTime();

        once = false;
    }

    validateData(msg);

    Timestamp msgTime = msg->getTimeStamp();
    Timestamp msgTimeNew = startTime + (msgTime - msgTime0);

    long sleepTime = (long)(msgTimeNew - DxcData::currentTime());

    DxcData* newMsg = cloneMessage(msg);
    newMsg->setTimeStamp(msgTimeNew);

    if (!abbrev && sleepTime > 0)
    {
        msleep(sleepTime);
    }
    pthread_mutex_lock(&connectorMutex);
    DxcConnector *tmp = connector;
    if (connector)
    {
        if (!connector->sendMessage(newMsg))
        {
            transFailedMessages += 1;
        }
    }
    pthread_mutex_unlock(&connectorMutex);

    delete newMsg;
}

void loadFreeSensorList()
{
    if (!sd)
        return;
    freeSensors = new set<string>();
    const map<string, double>* costs = sd->getSensorsAndCosts();

    map<string, double>::const_iterator i;
    for (i = costs->begin(); i != costs->end(); i++)
    {
        if (i->second == 0)
        {
            freeSensors->insert(i->first);
        }
    }
}

void printUsage(const string& cmd)
{
    cout << "usage: " << cmd << " [-ha] [-v system_name] [-t wait_time]";
    cout << " [-o output_file] input_file" << endl;
    cout << "       [-h]:             displays this message" << endl;
    cout << "       [-a]:             run in abbreviated mode, outputting data" << endl
         << "                         as quickly as possible (without sleep time" << endl
         << "                         between messages)." << endl;
    cout << "       [-v system_name]: validates scenario file against specified" << endl
         << "                         system's catalog." << endl;
    cout << "       [-t wait_time]:   specifies how long " + cmd + " should" << endl
         << "                         wait after scenario end before exiting." << endl;
    cout << "       [-o output_file]: file to store scenario results." << endl;
    cout << "       input_file:       scenario to run (must be in proper format)."
         << endl << endl;
}

void removeQuotes(string& str)
{
    size_t n = str.find_first_of("'\"");
    while (n != string::npos)
    {
        str.erase(n, 1);
        n = str.find_first_of("'\"", n);
    }
}


int main(int argc, char **argv)
{
    try
    {
        SDSCallback sdsc;

        pthreadsInit();
        pthread_mutex_init(&connectorMutex, NULL);
        pthread_mutex_init(&lastSensorsMutex, NULL);
        pthread_mutex_init(&daMutex, NULL);
        pthread_cond_init(&daCond, NULL);

        bool SLMode = false;
        long waitTime = 0;
        if (argc < 2)
        {
            printUsage(argv[0]);
            throw DxcException("Too few arguments, exiting.");
        }

        int pos = 1;

        while (pos < argc && argv[pos][0] == '-')
        {
            switch (argv[pos][1])
            {
            case 'h':
                printUsage(argv[0]);
                pos++;
                break;
            case 'a':
                abbrev = true;
                pos++;
                break;
            case 'v':
                sd = SystemCatalog::getInstance(argv[++pos]);
                pos++;
                break;
            case 't':
            {
                waitTime = -1;
                istringstream iss(argv[++pos]);
                try
                {
                    iss >> waitTime;
                    pos++;
                }
                catch (...)
                {
                    throw DxcException("-t: unable to parse time.");
                }
                if (waitTime == -1)
                    throw DxcException("-t: unable to parse time.");
                break;
            }
            case 's':
            {
                SLMode = true;
                pos++;
                break;
            }
            case 'o':
            {
                output = new ofstream(argv[pos + 1]);
                pos += 2;
                break;
            }
            default:
                throw DxcException(string(argv[pos]) + ": unknown option.");
            }
        }

        if (SLMode)
        {
            waitTime = 0;
            output = NULL;
        }
        else
        {
            if (!output || !output->good())
            {
                cout << "Unable to open results file for writing. "
                     << "Please specify a valid output file with [-o]." << endl;
                exit(1);
            }
            else
            {
                cout << "Writing to file." << endl;
            }
        }

        if (pos >= argc)
        {
            printUsage(argv[0]);
            throw DxcException("No input file specified.");
        }

        string filename(argv[pos]);
        removeQuotes(filename);

        ifstream tmpStream(filename.c_str(), ifstream::in);
        if (!tmpStream.good())
            throw DxcException("SDS: " + filename +
                               ": unable to open file");
        tmpStream.close();

        DxcConnector *slc = NULL, *src = NULL;
        try
        {
            if (!SLMode)
            {
                DxcConnector* slc = ConnectorFactory::getScenarioLoaderConnector(NULL);
                msleep(500);
            }

            connector = ConnectorFactory::getSDSConnector(&sdsc);

            if (!SLMode)
            {
                msleep(500);
                src = ConnectorFactory::getScenarioRecorderConnector(&sdsc);
            }
        }
        catch (...)
        {
            cerr << "SDS: ERROR: Unable to connect. Make sure SL is running." << endl;
            return 1;
        }

        MessageDriver *scenarioDriver = new MessageDriver();
        scenarioDriver->setMessageHook(sendMessage);
        if (!SLMode)
            cout << "Awaiting DA..." << endl;

        pthread_mutex_lock(&daMutex);
        pthread_cond_wait(&daCond, &daMutex);
        pthread_mutex_unlock(&daMutex);

        loadFreeSensorList();

        cout << "SDS: INFO: Parsing scenario file: " << filename << endl;
        if (abbrev)
        {
            cout << "SDS: INFO: Running in abbreviated mode." << endl;
        }
        cout.flush();
        scenarioDriver->parseFile(filename);
        if (transFailedMessages > 0)
        {
            cerr << "SDS: WARNING: Couldn't send " << transFailedMessages << " sensor messages to DA." << endl;
        }
        delete scenarioDriver;

        ScenarioStatusData ssd(ScenarioStatusData::SDS_ENDED);


        pthread_mutex_lock(&connectorMutex);
        if (connector)
        {
            if (!connector->sendMessage(&ssd))
            {
                transFailedMessages += 1;
            }
        }

        if (!SLMode)
            cout << "Scenario ended." << endl;

        pthread_mutex_unlock(&connectorMutex);
        pthread_mutex_lock(&daMutex);

        if (waitTime > 0)
        {
            timeval systemTime;
            timespec t;
            gettimeofday(&systemTime, NULL);
            t.tv_sec = systemTime.tv_sec + (waitTime);
            pthread_cond_timedwait(&daCond, &daMutex, &t);
        }
        else
            pthread_cond_wait(&daCond, &daMutex);

        pthread_mutex_unlock(&daMutex);
        pthread_mutex_lock(&connectorMutex);

        delete slc, connector, src;
        connector = NULL;
        pthread_mutex_unlock(&connectorMutex);

    }
    catch (DxcException ex)
    {
        try
        {
            cout << ex.what() << endl;

            pthread_mutex_lock(&connectorMutex);
            if (connector)
            {
                ErrorData ed(DxcData::currentTime(), ex.what());
                ScenarioStatusData ssd(ScenarioStatusData::SDS_ENDED);
                connector->sendMessage(&ed);
                connector->sendMessage(&ssd);

                if (output)
                    *output << &ed << &ssd;

                delete connector;
                connector = NULL;
            }
            pthread_mutex_unlock(&connectorMutex);
        }
        catch (...)
        {
            cerr << "SDS: ERROR: Unable to send error. " << endl;
        }
    }
    catch (...)
    {
        cerr << "SDS: ERROR: Unknown exception." << endl;
    }

    return 0;
}
